#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .artist import Artist
from .album import Album
from .track import Track
from .playlist import Playlist
from .genre import Genre

from .section import Section
from .message import Message

from .searchResults import SearchResults
from .musicFromPost import MusicFromPost


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.
from vkmusix.config import playlistsOwnerId


class GetMusicFromPost:
    from typing import Union

    from vkmusix.aio import async_
    from vkmusix.types import MusicFromPost

    @async_
    async def getMusicFromPost(self, postId: int, ownerId: int = None) -> Union[MusicFromPost, None]:
        """
        Получает музыку из поста.

        `Пример использования`:

        music = client.getMusicFromPost(
            postId=34450119,
            ownerId=-28905875,
        )

        print(music)

        :param postId: идентификатор поста. (``int``)
        :param ownerId: идентификатор owner'а (пользователь или группа). По умолчанию залогиненный пользователь. (``int``, `optional`)
        :return: `При успехе`: музыка из поста (``types.MusicFromPost``). `Если пост не найден или музыка отсутствует`: ``None``.
        """

        import asyncio

        import re

        from vkmusix.types import Album, Track, Playlist, MusicFromPost

        if not ownerId:
            ownerId = await self._getMyId()

        post = (await self._req(
            "wall.getById",
            {
                "posts": f"{ownerId}_{postId}",

            },
        )).get("items")

        if not post:
            return

        post = post[0]
        attachments = post.get("attachments")

        if not attachments:
            return

        tracks = list()
        playlists = list()
        playlistTasks = list()

        for attachment in attachments:
            type = attachment.get("type")

            if type == "audio":
                tracks.append(attachment.get("audio"))

            elif type == "link":
                link = attachment.get("link")
                url = link.get("url")

                if url.startswith("https://m.vk.com/audio?act=audio_playlist"):
                    match = re.search(r"audio_playlist(-?\d+)_(\d+)", url)
                    if not match:
                        continue

                    ownerId, playlistId = match.groups()

                    playlistTask = self.getPlaylist(
                        playlistId,
                        ownerId,
                    )

                    playlistTasks.append(playlistTask)

        if playlistTasks:
            playlists = await asyncio.gather(*playlistTasks)

            if playlists:
                playlists = [playlist for playlist in playlists if playlist]

        albums = [album for album in playlists if isinstance(album, Album)] if playlists else None
        playlists = [playlist for playlist in playlists if isinstance(playlist, Playlist)] if playlists else None

        return MusicFromPost(
            albums if albums else None,
            self._finalizeResponse(tracks, Track) if tracks else None,
            playlists if playlists else None,
        ) if any((albums, tracks, playlists)) else None

    get_music_from_post = getMusicFromPost


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetTracksFromChat:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Message

    @async_
    async def getTracksFromChat(self, chatId: int = None, groupId: int = None) -> Union[List[Message], None]:
        """
        Получает сообщения с треками из чата.

        `Пример использования`:

        messages = client.getTracksFromChat()

        print(messages)

        :param chatId: идентификатор чата. По умолчанию «Избранное». (``int``, `optional`)
        :param groupId: идентификатор группы, в которой находится чат. Только для личных сообщений группы. (``int``, `optional`)
        :return: `При успехе`: сообщения с треками (``list[types.Message]``). `Если чат не найден или сообщения с треками отсутствуют`: ``None``.
        """

        from vkmusix.types import Track, Message

        if not chatId:
            chatId = await self._getMyId()

        messages = (await self._req(
            "messages.getHistoryAttachments",
            {
                "media_type": "audio",
                "count": 200,
                "peer_id": chatId,
                **({
                       "group_id": groupId,
                } if groupId else dict()),
            },
            version=5.241,
        )).get("items")

        if not messages:
            return

        messages = messages[::-1]

        newMessages = dict()

        for idx, message in enumerate(messages):
            messageId = message.get("message_id")
            messageInfo = newMessages.get(messageId)

            track = self._finalizeResponse(
                message.get("attachment").get("audio"),
                Track,
            )

            if messageInfo:
                messageInfo.get("tracks").append(track)

            else:
                newMessages[messageId] = {
                    "tracks": [track],
                    "date": message.get("date"),
                    "from_id": message.get("from_id"),
                    "chat_id": chatId,
                    "message_id": messageId,
                }

        return [
            Message(message)
            for message in newMessages.values()
        ]

    get_tracks_from_chat = getTracksFromChat


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetSection:
    from typing import Union

    from vkmusix.aio import async_
    from vkmusix.types import Section

    @async_
    async def getSection(self, sectionId: str, offset: str = None) -> Union[Section, None]:
        """
        Получает треки из раздела музыки.

        `Пример использования`:

        section = client.getSection(
            sectionId="PUlQVA8GR0R3W0tMF1kSOSceDR9aRzQEKgQKHRcYSV5kUUREDQ1bU35cXFoXAFtEfEZYWhcBSVxkDAwYUEYKCmRHS0IXDlpKZFpcVA8FR0R0XUtMGAZTX3ZeUUEASQ",
        )

        print(section)

        :param sectionId: идентификатор раздела музыки. (``str``)
        :param offset: уникальное значение, содержащееся в атрибуте ``nextOffset`` объекта класса ``types.Section``. Необходимо для получения следующих результатов в разделе. (``str``, `optional`)
        :return: `При успехе`: информация о разделе музыки (``types.Section``). `Если раздел не найден`: ``None``.
        """

        from vkmusix.types import Section

        section = await self._req(
            "catalog.getSection",
            {
                "section_id": sectionId,
                **({
                    "start_from": offset,
                } if offset else dict()),
            },
        )

        if not section:
            return

        return self._finalizeResponse(section, Section)

    get_section = getSection


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .base import Base

class MusicFromPost(Base):
    """
    Класс, представляющий музыку из поста.

    Атрибуты:
        albums (list[types.Album], optional): альбомы из поста.

        tracks (list[types.Track], optional): треки из поста.

        playlists (list[types.Playlist], optional): плейлисты из поста.
    """

    from typing import Union, List

    from vkmusix.types import Album, Track, Playlist

    def __init__(self, albums: Union[List[Album], None] = None, tracks: Union[List[Track], None] = None, playlists: Union[List[Playlist], None] = None) -> None:
        self.albums = albums
        self.tracks = tracks
        self.playlists = playlists


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .getPlaylist import GetPlaylist
from .getPlaylistTracks import GetPlaylistTracks
from .removeAllTracksFromPlaylist import RemoveAllTracksFromPlaylist

from .createPlaylist import CreatePlaylist
from .addPlaylist import AddPlaylist
from .removePlaylist import RemovePlaylist
from .editPlaylist import EditPlaylist
from .copyPlaylist import CopyPlaylist
from ._editPlaylistPhoto import _EditPlaylistPhoto

class Playlists(
    GetPlaylist,
    GetPlaylistTracks,
    RemoveAllTracksFromPlaylist,

    AddPlaylist,
    RemovePlaylist,
    CreatePlaylist,
    EditPlaylist,
    CopyPlaylist,
    _EditPlaylistPhoto,
):
    pass


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Get:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Track

    @async_
    async def get(self, ownerIds: Union[List[int], int], trackIds: Union[List[int], int], includeLyrics: bool = False) -> Union[List[Track], Track, None]:
        """
        Получает информацию о треках.

        `Пример использования`:

        track = client.get(
            ownerIds=-2001471901,
            trackIds=123471901,
            includeLyrics=True,
        )

        print(track)

        :param ownerIds: идентификаторы владельцев треков. Максимум 343. (``Union[list[int], int]``)
        :param trackIds: идентификаторы треков. Максимум 343. (``Union[list[int], int]``)
        :param includeLyrics: флаг, указывающий, небходимо ли также получить текст. (``bool``, `optional`)
        :return: `Если треков несколько`: информация о треках (``list[types.Track]``). `Если трек один`: информация о треке (``types.Track``). `При успехе`: информация о треке (``types.Track``). `Если трек не найден`: ``None``.
        """

        from asyncio import gather

        from vkmusix.types import Track

        if type(ownerIds) != type(trackIds):
            self._raiseError("ownerIdsAndTrackIdsTypeDifferent")

        elif isinstance(ownerIds, list) and isinstance(trackIds, list) and len(ownerIds) != len(trackIds):
            self._raiseError("ownerIdsAndTrackIdsLenDifferent")

        wasList = isinstance(ownerIds, list)

        if not all((isinstance(ownerIds, list), isinstance(trackIds, list))):
            ownerIds = [ownerIds]
            trackIds = [trackIds]

        ids =  [
            f"{ownerId}_{trackId}"
            for ownerId, trackId in zip(
                ownerIds,
                trackIds,
            )
        ]

        tasks = [
            self._req(
                "getById",
                {
                    "audios": ",".join(ids),
                },
            ),
        ]

        if includeLyrics:
            for id in ids:
                ownerId, trackId = id.split("_")
                tasks.append(
                    self.getLyrics(
                        ownerId,
                        trackId,
                    ),
                )

        responses = await gather(*tasks)

        tracks = responses[0]
        if not tracks:
            return

        if not isinstance(tracks, list):
            tracks = [tracks]

        if includeLyrics:
            lyrics = responses[1:]

            for idx, track in enumerate(tracks):
                track["lyrics"] = lyrics[idx] if lyrics else None

        return self._finalizeResponse(tracks if wasList else tracks[0], Track)


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetPlaylistTracks:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Track

    @async_
    async def getPlaylistTracks(self, playlistId: int, ownerId: int = None, isLarge: bool = False) -> Union[List[Track], None]:
        """
        Получает треки плейлиста или альбома.

        `Пример использования`:

        tracks = client.getPlaylistTracks(
            playlistId=19201020,
            ownerId=-2000201020,
        )

        print(tracks)

        :param playlistId: идентификатор плейлиста или альбома. (``int``)
        :param ownerId: идентификатор владельца плейлиста или альбома (пользователь или группа). (``int``, `optional`)
        :param isLarge: флаг, указывающий, содержит ли плейлист или альбом более 1000 треков. По умолчанию ``False``. Если ``True``, будет получена ограниченная информация о всех треках, если ``False`` — только первые 1000 треков, но с полной информацией. Установите ``None``, чтобы библиотека определила автоматически. Игнорируется для приватных плейлистов. (``bool``, optional)
        :return: `При успехе`: треки плейлиста или альбома (``list[types.Track]``). `Если плейлист или альбом не найден или треки отсутствуют`: ``None``.
        """

        from vkmusix.config import VK, headers

        if not ownerId:
            ownerId = await self._getMyId()

        if isLarge is None:
            playlist = await self.getPlaylist(
                playlistId,
                ownerId,
            )

            if not playlist:
                return

            isLarge = playlist.trackCount and playlist.trackCount > 1000

        if not isLarge:
            tracks = await self._client.req(
                f"{VK}music/playlist/{ownerId}_{playlistId}",
                headers=headers,
                responseType="response",
            )

            statusCode = tracks.status_code

            if statusCode == 404:
                return

            while statusCode == 302:
                tracks = await self._client.req(
                    f'{VK}{tracks.headers.get("Location")}',
                    headers=headers,
                    responseType="response",
                )
                statusCode = tracks.status_code

            tracks = await self._parseWebTracks(tracks.text)

        else:
            tracks = None

        if not tracks:
            tracks = (await self._req(
                "getAudioIdsBySource",
                {
                    "source": "playlist",
                    "entity_id": f"{ownerId}_{playlistId}",
                },
            )).get("audios")

            tracks = await self._parseAPITracks(tracks)

        return tracks if tracks else None

    get_playlist_tracks = getPlaylistTracks


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class RemoveAllTracksFromPlaylist:
    from vkmusix.aio import async_

    @async_
    async def removeAllTracksFromPlaylist(self, playlistId: int = None, groupId: int = None) -> bool:
        """
        Удаляет все треки из плейлиста пользователя или группы. Метод не работает для плейлистов, привязанных к чату.

        `Пример использования`:

        result = client.removeAllTracksFromPlaylist(
            playlistId=19201020,
        )

        print(result)

        :param playlistId: идентификатор плейлиста. (``int``, `optional`)
        :param groupId: идентификатор группы, из плейлиста которой необходимо удалить треки. (``int``, `optional`)
        :return: `При успехе`: ``True``. `Если плейлист не найден, треки отсутствуют или их не удалось удалить`: ``False``.
        """

        tracks = await self.getPlaylistTracks(
            playlistId,
            groupId,
        )

        if not tracks:
            return False

        ownerIds = [track.ownerId for track in tracks]
        trackIds = [track.trackId for track in tracks]

        result = await self.remove(
            ownerIds,
            trackIds,
            playlistId,
            groupId,
            False,
        )

        return bool(result[0])


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class _ParseAPITracks:
    from typing import Union, List, Dict

    from vkmusix.aio import async_
    from vkmusix.types import Track

    @async_
    async def _parseAPITracks(self, tracks: Union[List[Dict[str, str]], None]) -> Union[List[Track], None]:
        from vkmusix.types import Track

        if not tracks:
            return

        if not isinstance(tracks, list):
            tracks = [tracks]

        for idx, track in enumerate(tracks):
            id = track.get("audio_id")

            if id.count("_") == 3:
                id = id[:id.rfind("_")]

            ownerId, trackId = id.split("_")[:2]
            tracks[idx] = {
                "owner_id": int(ownerId),
                "track_id": int(trackId),
            }

        return self._finalizeResponse(tracks, Track)


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class GetSections:
    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Section

    @async_
    async def getSections(self, ownerId: int = None) -> Union[List[Section], None]:
        """
        Получает разделы музыки owner'а (пользователь или группа).

        `Пример использования`:

        sections = client.getSections()

        print(sections)

        :param ownerId: идентификатор owner'а (пользователь или группа). По умолчанию залогиненный пользователь. (``int``, `optional`)
        :return: `При успехе`: разделы музыки (``List[types.Section]``). `Если owner (пользователь или группа) не найден`: ``None``.
        """

        from vkmusix.types import Section

        if not ownerId:
            ownerId = await self._getMyId()

        catalog = await self._req(
            "catalog.getAudio",
            {
                "owner_id": ownerId,
            },
        )

        catalog = catalog.get("catalog")
        sections = catalog.get("sections")

        if not sections:
            return

        return self._finalizeResponse(sections, Section)

    get_sections = getSections


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Remove:
    from typing import Union, List

    from vkmusix.aio import async_

    @async_
    async def remove(self, ownerIds: Union[List[int], int], trackIds: Union[List[int], int], playlistId: int = None, groupId: int = None, validateIds: bool = True) -> Union[List[bool], bool]:
        """
        Удаляет треки из музыки или плейлиста пользователя или группы.

        `Пример использования`:

        result = client.remove(
            ownerIds=-2001471901,
            trackIds=123471901,
        )

        print(result)

        :param ownerIds: идентификаторы владельцев треков. (``Union[list[int], int]``)
        :param trackIds: идентификаторы треков. (``Union[list[int], int]``)
        :param playlistId: идентификатор плейлиста, из которого необходимо удалить треки. Метод не работает для плейлистов, привязанных к чату. (``int``, `optional`)
        :param groupId: идентификатор группы, из музыки или плейлиста которой необходимо удалить треки. (``int``, `optional`)
        :param validateIds: флаг, указывающий, необходимо ли перепроверить треки на наличие в музыке или плейлисте. По умолчанию ``True``. Установите на ``False``, если вы получили треки через ``client.getTracks()`` или ``client.getSection()`` (при удалении из музыки) или ``client.getPlaylistTracks()`` (при удалении из плейлиста). (``bool``, `optional`)
        :return: `Если треков несколько`: статусы удаления треков (``list[bool]``). `Если трек один`: статус удаления трека (``bool``). `При успехе`: ``True``. `Если трек не удалось удалить`: ``False``.
        """

        import asyncio

        from itertools import islice

        from vkmusix.errors import AccessDenied

        if type(ownerIds) != type(trackIds):
            self._raiseError("ownerIdsAndTrackIdsTypeDifferent")

        elif isinstance(ownerIds, list) and isinstance(trackIds, list) and len(ownerIds) != len(trackIds):
            self._raiseError("ownerIdsAndTrackIdsLenDifferent")

        wasList = isinstance(ownerIds, list)

        if not all((isinstance(ownerIds, list), isinstance(trackIds, list))):
            ownerIds = [ownerIds]
            trackIds = [trackIds]

        if not groupId:
            groupId = await self._getMyId()

        if validateIds:
            if playlistId:
                existTracks = await self.getPlaylistTracks(
                    playlistId,
                    groupId,
                )

                if not existTracks:
                    return [False] * len(ownerIds)

            else:
                sections = await self.getSections(
                    groupId,
                )

                if not sections:
                    return [False] * len(ownerIds)

                subsections = (await sections[0].get()).subsections

                existTracks = await subsections[0].getTracks()

                if not existTracks:
                    return [False] * len(ownerIds)

                existOwnerIds = [track.ownerId for track in existTracks]
                existTrackIds = [track.trackId for track in existTracks]

                batchSize = 343
                tasks = [
                    self.get(
                        existOwnerIds[i:i + batchSize],
                        existTrackIds[i:i + batchSize]
                    )
                    for i in range(0, len(existTracks), batchSize)
                ]

                batches = await asyncio.gather(*tasks)

                existTracks = [
                    track
                    for batch in batches if batch
                    for track in batch
                ]

            for idx, (ownerId, trackId) in enumerate(zip(ownerIds, trackIds)):
                track = await self.get(ownerId, trackId)

                trackFullTitle = track.fullTitle
                trackArtists = str(track.artists) or track.artist

                for existTrack in existTracks:
                    existTrackFullTitle = existTrack.fullTitle
                    existTrackArtists = str(existTrack.artists) or existTrack.artist

                    if existTrackFullTitle == trackFullTitle and existTrackArtists == trackArtists:
                        ownerIds[idx] = existTrack.ownerId
                        trackIds[idx] = existTrack.trackId
                        break

        results = list()

        if playlistId:
            def chunks(iterable):
                iterator = iter(iterable)
                for first in iterator:
                    yield [first] + list(islice(iterator, 99))

            for trackChunk in chunks(zip(ownerIds, trackIds)):
                ids = ",".join(
                    f"{ownerId}_{trackId}"
                    for ownerId, trackId in trackChunk
                )

                try:
                    response = await self._req(
                        "removeFromPlaylist",
                        {
                            "owner_id": groupId,
                            "audio_ids": ids,
                            "playlist_id": playlistId,
                        },
                    )

                    results.append([bool(response)] * len(trackChunk))

                except AccessDenied:
                    results.extend([False] * len(trackChunk))

        else:
            for ownerId, trackId in zip(ownerIds, trackIds):
                try:
                    response = await self._req(
                        "delete",
                        {
                            "owner_id": ownerId,
                            "audio_id": trackId,
                            "group_id": groupId,
                        },
                    )
                    results.append(bool(response))

                except AccessDenied:
                    results.append(False)

        return results if wasList else results[0]


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .base import Base

class Message(Base):
    """
    Класс, представляющий сообщение с треками.

    Атрибуты:
        tracks (list[types.Track]): треки сообщения.

        date (datetime): дата и время отправки сообщения.

        fromId (int): идентификатор автора сообщения.

        chatId (int): идентификатор чата, в котором отправлено сообщение.

        id (int): идентификатор сообщения.

        raw (dict): необработанные данные, полученные от ВКонтакте.
    """

    def __init__(self, message: dict) -> None:
        from vkmusix.utils import unixToDatetime

        self.tracks = message.get("tracks")

        self.date = unixToDatetime(message.get("date"))

        self.fromId = message.get("from_id")

        self.chatId = message.get("chat_id")

        self.id = message.get("message_id")

        self.raw = message


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Add:
    from typing import Union, List

    from vkmusix.aio import async_

    @async_
    async def add(self, ownerIds: Union[List[int], int], trackIds: Union[List[int], int], playlistId: int = None, groupId: int = None) -> Union[List[bool], bool]:
        """
        Добавляет треки в музыку или плейлист пользователя или группы.

        `Пример использования`:

        result = client.add(
            ownerIds=-2001471901,
            trackIds=123471901,
        )

        print(result)

        :param ownerIds: идентификаторы владельцев треков. (``Union[list[int], int]``)
        :param trackIds: идентификаторы треков. (``Union[list[int], int]``)
        :param playlistId: идентификатор плейлиста, в который необходимо добавить треки. (``int``, `optional`)
        :param groupId: идентификатор группы, в музыку или плейлист которой необходимо добавить треки. (``int``, `optional`)
        :return: `Если треков несколько`: статусы добавления треков (``list[bool]``). `Если трек один`: статус добавления трека (``bool``). `При успехе`: ``True``. `Если трек не удалось добавить`: ``False``.
        """

        from itertools import islice

        from vkmusix.errors import AccessDenied

        if type(ownerIds) != type(trackIds):
            self._raiseError("ownerIdsAndTrackIdsTypeDifferent")

        elif isinstance(ownerIds, list) and isinstance(trackIds, list) and len(ownerIds) != len(trackIds):
            self._raiseError("ownerIdsAndTrackIdsLenDifferent")

        wasList = isinstance(ownerIds, list)

        if not all((isinstance(ownerIds, list), isinstance(trackIds, list))):
            ownerIds = [ownerIds]
            trackIds = [trackIds]

        if not groupId:
            groupId = await self._getMyId()

        method = "add"
        if playlistId:
            method += "ToPlaylist"

        results = list()

        if playlistId:
            def chunks(iterable):
                iterator = iter(iterable)
                for first in iterator:
                    yield [first] + list(islice(iterator, 49))

            for trackChunk in chunks(zip(ownerIds, trackIds)):
                ids = ",".join(
                    f"{ownerId}_{trackId}"
                    for ownerId, trackId in trackChunk
                )

                try:
                    response = await self._req(
                        method,
                        {
                            "owner_id": groupId,
                            "audio_ids": ids,
                            "playlist_id": playlistId,
                        },
                    )

                    results.append([bool(response)] * len(trackChunk))

                except AccessDenied:
                    print(1)
                    results.extend([False] * len(trackChunk))

        else:
            for ownerId, trackId in zip(ownerIds, trackIds):
                try:
                    response = await self._req(
                        method,
                        {
                            "owner_id": ownerId,
                            "audio_id": trackId,
                            "group_id": groupId,
                        },
                    )
                    results.append(bool(response))

                except AccessDenied:
                    results.append(False)

        return results if wasList else results[0]


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .getTrackCount import GetTrackCount
from .getTracks import GetTracks

from .getSections import GetSections
from .getSection import GetSection

from .reorder import Reorder

from .getPlaylists import GetPlaylists
from .getAllPlaylists import GetAllPlaylists

from .getBroadcast import GetBroadcast
from .setBroadcast import SetBroadcast

from .getTracksFromWall import GetTracksFromWall
from .getMusicFromPost import GetMusicFromPost

from .followOwner import FollowOwner
from .unfollowOwner import UnfollowOwner

class Owners(
    GetTrackCount,
    GetTracks,

    GetSections,
    GetSection,

    Reorder,

    GetPlaylists,
    GetAllPlaylists,

    GetBroadcast,
    SetBroadcast,

    GetTracksFromWall,
    GetMusicFromPost,

    FollowOwner,
    UnfollowOwner,
):
    pass


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

from .base import Base

class Section(Base):
    """
    Класс, представляющий раздел музыки owner'а (пользователь или группа).

    Атрибуты:
        title (str): название раздела.

        subsections (list[types.Section], optional): подразделы.

        tracks (list[types.Track], optional): треки раздела.

        playlists (list[types.Playlist], optional): плейлисты раздела.

        recommendedPlaylists (list[types.Playlist], optional): рекомендованные плейлисты.

        nextOffset (str, optional): значение оффсета для получения следующих результатов. Отсутствует, если результатов больше нету.

        id (str): идентификатор раздела.

        url (str): ссылка на раздел в web-версии ВКонтакте.

        raw (dict): необработанные данные, полученные от ВКонтакте.
    """

    from typing import Union, List

    from vkmusix.aio import async_
    from vkmusix.types import Track

    def __init__(self, section: dict, client: "Client" = None) -> None:
        from vkmusix.types import Track, Playlist

        super().__init__(client)

        sectionInfo = section.get("section", section)

        self.title = sectionInfo.get("title")

        subsections = sectionInfo.get("blocks")

        self.subsections = self._client._finalizeResponse(
            [subsection for subsection in subsections if subsection.get("data_type") not in ["none", "audio_stream_mixes"]],
            Section,
        ) if subsections else None

        tracks = section.get("audios")

        if not tracks:
            ids = section.get("audios_ids")

            if ids:
                tracks = [
                    {
                        "owner_id": ownerId,
                        "track_id": trackId,
                    }
                    for id in ids
                    for ownerId, trackId in [id.split("_")]
                ]

        self.tracks = self._client._finalizeResponse(
            tracks,
            Track,
        )

        playlists = section.get("playlists")

        if not playlists:
            ids = section.get("playlists_ids")

            if ids:
                playlists = [
                    {
                        "owner_id": ownerId,
                        "playlist_id": playlistId,
                    }
                    for id in ids
                    for ownerId, playlistId in [id.split("_")]
                ]

        self.playlists = self._client._finalizeResponse(
            playlists,
            Playlist,
        )

        self.recommendedPlaylists = self._client._finalizeResponse(
            section.get("recommended_playlists"),
            Playlist,
        )

        self.nextOffset = sectionInfo.get("next_from")

        self.id = sectionInfo.get("id")
        self.url = sectionInfo.get("url")

        if self.subsections:
            if len(self.subsections) == 1 and self.subsections[0].id == self.id:
                self.subsections = None

            else:
                self.subsections = [subsection for subsection in self.subsections if subsection.id != self.id]

        self.raw = section


    @async_
    async def get(self, offset: str = None) -> "Section":
        """
        Получает информацию о разделе музыки.

        `Пример использования`:

        section = section.get()

        print(section)

        :param offset: уникальное значение, содержащееся в атрибуте ``nextOffset`` объекта класса ``types.Section``. Необходимо для получения следующих результатов в разделе. (``str``, `optional`)
        :return: `При успехе`: информация о разделе музыки (``types.Section``). `Если раздел не найден`: ``None``.
        """

        return await self._client.getSection(
            self.id,
            offset,
        )


    @async_
    async def getTracks(self) -> Union[List[Track], None]:
        """
        Получает треки из раздела музыки.

        `Пример использования`:

        tracks = section.getTracks()

        print(tracks)

        :return: `При успехе`: треки (``list[types.Track]``). `Если раздел не найден или треки отсутствуют`: ``None``.
        """

        return await self._client.getTracks(
            self.id,
        )

    get_tracks = getTracks


#  VKMusix — VK Music API Client Library for Python
#  Copyright (C) 2024—present to4no4sv <https://github.com/to4no4sv/VKMusix>
#
#  This file is part of VKMusix.
#
#  VKMusix is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  VKMusix is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with VKMusix. If not, see <http://www.gnu.org/licenses/>.

class Search:
    from typing import Union

    from vkmusix.aio import async_
    from vkmusix.types import SearchResults

    @async_
    async def search(self, query: str, limit: int = None, offset: int = None) -> Union[SearchResults, None]:
        """
        Ищет артистов, альбомы, треки и плейлисты.

        `Пример использования`:

        searchResults = client.search(
            query="Маленький ярче",
            limit=10,
        )

        print(searchResults)

        :param query: поисковой запрос. (``str``)
        :param limit: лимит объектов каждого типа. (``int``, `optional`)
        :param offset: сколько объектов каждого типа пропустить. (``int``, `optional`)
        :return: `При успехе`: результаты поиска (``types.SearchResults``). `Если ничего не найдено`: ``None``.
        """

        from vkmusix.types import Artist, Album, Track, Playlist

        return await self._search(
            "searchMain",
            (query, limit, offset),
            [Artist, Album, Track, Playlist],
        )